@@ -4,7 +4,7 @@ RDF::Endpoint - A SPARQL Protocol Endpoint implementation
 
 =head1 VERSION
 
-This document describes RDF::Endpoint version 0.04.
+This document describes RDF::Endpoint version 0.06.
 
 =head1 SYNOPSIS
 
@@ -55,6 +55,16 @@ included for execution. The boolean values 'default' and 'named_graphs' indicate
 that the respective SPARQL dataset graphs should be described by the service
 description.
 
+=item html
+
+An associative array (hash) containing details on how results should be
+serialized when the output media type is HTML. The boolean value 'resource_links'
+specifies whether URI values should be serialized as HTML anchors (links).
+The boolean value 'embed_images' specifies whether URI values that are typed as
+foaf:Image should be serialized as HTML images. If 'embed_images' is true, the
+integer value 'image_width' specifies the image width to be used in the HTML
+markup (letting the image height scale appropriately).
+
 =back
 
 =back
@@ -93,11 +103,12 @@ package RDF::Endpoint;
 use 5.008;
 use strict;
 use warnings;
-our $VERSION	= '0.04';
+our $VERSION	= '0.06';
 
 use RDF::Query 2.905;
 use RDF::Trine 0.134 qw(statement iri blank literal);
 
+use JSON;
 use Encode;
 use File::Spec;
 use Data::Dumper;
@@ -109,11 +120,13 @@ use Scalar::Util qw(blessed refaddr);
 use File::ShareDir qw(dist_dir);
 use HTTP::Negotiate qw(choose);
 use RDF::Trine::Namespace qw(rdf xsd);
-use RDF::RDFa::Generator 0.101;
+use RDF::RDFa::Generator 0.102;
 use IO::Compress::Gzip qw(gzip);
 use HTML::HTML5::Parser;
 use HTML::HTML5::Writer qw(DOCTYPE_XHTML_RDFA);
 use Hash::Merge::Simple qw/ merge /;
+use Fcntl qw(:flock SEEK_END);
+use Carp qw(croak);
 
 
 my $NAMESPACES	= {
@@ -127,6 +140,8 @@ my $NAMESPACES	= {
 	ldodds		=> 'java:com.ldodds.sparql.',
 	fn			=> 'http://www.w3.org/2005/xpath-functions#',
 	sparql		=> 'http://www.w3.org/ns/sparql#',
+	vann		=> 'http://purl.org/vocab/vann/',
+	sde			=> 'http://kasei.us/ns/service-description-extension#',
 };
 
 =item C<< new ( \%conf ) >>
@@ -151,9 +166,15 @@ sub new {
 		my $store	= RDF::Trine::Store->new( $config->{store} );
 		$model		= RDF::Trine::Model->new( $store );
 	}
+	
 	unless ($config->{endpoint}) {
 		$config->{endpoint}	= { %$config };
 	}
+	
+	if ($config->{endpoint}{load_data} and $config->{endpoint}{update}) {
+		die "The load_data and update configuration options cannot be specified together.";
+	}
+	
 	my $self	= bless( {
 		conf		=> $config,
 		model		=> $model,
@@ -173,20 +194,26 @@ an appropriate Plack::Response object.
 sub run {
 	my $self	= shift;
 	my $req		= shift;
+	
 	my $config	= $self->{conf};
-	my $endpoint_path = $config->{endpoint}->{endpoint_path} || '/sparql';
+	my $endpoint_path = $config->{endpoint}{endpoint_path} || '/sparql';
 	$config->{resource_links}	= 1 unless (exists $config->{resource_links});
 	my $model	= $self->{model};
 	
 	my $content;
 	my $response	= Plack::Response->new;
+
+	my $server = "RDF::Endpoint/$VERSION";
+	$server .= " " . $response->headers->header('Server') if defined($response->headers->header('Server'));
+	$response->headers->header('Server' => $server);
+
 	unless ($req->path eq $endpoint_path) {
 		my $path	= $req->path_info;
 		$path		=~ s#^/##;
 		my $dir		= $ENV{RDF_ENDPOINT_SHAREDIR} || eval { dist_dir('RDF-Endpoint') } || 'share';
 		my $file	= File::Spec->catfile($dir, 'www', $path);
 		if (-r $file) {
-			open( my $fh, '<', $file ) or die $!;
+			open( my $fh, '<', $file ) or croak $!;
 			$response->status(200);
 			$content	= $fh;
 		} else {
@@ -201,7 +228,6 @@ END
 		return $response;
 	}
 	
-	
 	my $headers	= $req->headers;
 	my $type	= $headers->header('Accept') || 'application/sparql-results+xml';
 	if (my $t = $req->param('media-type')) {
@@ -213,20 +239,62 @@ END
 	
 	my $sparql;
 	my $ct	= $req->header('Content-type');
-	if (defined($ct) and $ct eq 'application/sparql-query') {
+	if ($req->method !~ /^(GET|POST)$/i) {
+		my $method	= uc($req->method);
+		$content	= "Unexpected method $method (expecting GET or POST)";
+		$self->log_error( $req, $content );
+		$self->_set_response_error($req, $response, 405, {
+			title		=> 'Method not allowed',
+			describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/bad_http_method',
+		});
+		$response->header('Allow' => 'GET, POST');
+		goto CLEANUP;
+	} elsif (defined($ct) and $ct eq 'application/sparql-query') {
 		$sparql	= $req->content;
 	} elsif (defined($ct) and $ct eq 'application/sparql-update') {
 		if ($config->{endpoint}{update} and $req->method eq 'POST') {
 			$sparql	= $req->content;
 		}
 	} elsif ($req->param('query')) {
-		$sparql = $req->param('query');
+		my @sparql	= $req->param('query');
+		if (scalar(@sparql) > 1) {
+			$content	= "More than one query string submitted";
+			$self->log_error( $req, $content );
+			$self->_set_response_error($req, $response, 400, {
+				title		=> 'Multiple query strings not allowed',
+				describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/multiple_queries',
+			});
+			goto CLEANUP;
+		} else {
+			$sparql = $sparql[0];
+		}
 	} elsif ($req->param('update')) {
+		my @sparql	= $req->param('update');
+		if (scalar(@sparql) > 1) {
+			$content	= "More than one update string submitted";
+			$self->log_error( $req, $content );
+			$self->_set_response_error($req, $response, 400, {
+				title		=> 'Multiple update strings not allowed',
+				describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/multiple_updates',
+			});
+			goto CLEANUP;
+		}
+		
 		if ($config->{endpoint}{update} and $req->method eq 'POST') {
-			$sparql = $req->param('update');
+			$sparql = $sparql[0];
+		} elsif ($req->method ne 'POST') {
+			my $method	= $req->method;
+			$content	= "Update operations must use POST";
+			$self->log_error( $req, $content );
+			$self->_set_response_error($req, $response, 405, {
+				title		=> "$method Not Allowed for Update Operation",
+				describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/bad_http_method_update',
+			});
+			$response->header('Allow' => 'POST');
+			goto CLEANUP;
 		}
 	}
-
+	
 	my $ns = merge $config->{namespaces}, $NAMESPACES;
 
 	if ($sparql) {
@@ -234,16 +302,29 @@ END
 		$args{ update }		= 1 if ($config->{endpoint}{update} and $req->method eq 'POST');
 		$args{ load_data }	= 1 if ($config->{endpoint}{load_data});
 		
-		my @default	= $req->param('default-graph-uri');
-		my @named	= $req->param('named-graph-uri');
-		if (scalar(@default) or scalar(@named)) {
-			delete $args{ load_data };
-			$model	= RDF::Trine::Model->new( RDF::Trine::Store::Memory->new() );
-			foreach my $url (@named) {
-				RDF::Trine::Parser->parse_url_into_model( $url, $model, context => iri($url) );
+		{
+			my @default	= $req->param('default-graph-uri');
+			my @named	= $req->param('named-graph-uri');
+			if (scalar(@default) or scalar(@named)) {
+				delete $args{ load_data };
+				$model	= RDF::Trine::Model->new( RDF::Trine::Store::Memory->new() );
+				foreach my $url (@named) {
+					RDF::Trine::Parser->parse_url_into_model( $url, $model, context => iri($url) );
+				}
+				foreach my $url (@default) {
+					RDF::Trine::Parser->parse_url_into_model( $url, $model );
+				}
 			}
-			foreach my $url (@default) {
-				RDF::Trine::Parser->parse_url_into_model( $url, $model );
+		}
+		
+		my $protocol_specifies_update_dataset	= 0;
+		{
+			my @default	= $req->param('using-graph-uri');
+			my @named	= $req->param('using-named-graph-uri');
+			if (scalar(@named) or scalar(@default)) {
+				$protocol_specifies_update_dataset	= 1;
+				$model	= RDF::Trine::Model::Dataset->new( $model );
+				$model->push_dataset( default => \@default, named => \@named );
 			}
 		}
 		
@@ -258,7 +339,19 @@ END
 		
 		my $base	= $req->base;
 		my $query	= RDF::Query->new( $sparql, { lang => 'sparql11', base => $base, %args } );
+		$self->log_query( $req, $sparql );
 		if ($query) {
+			if ($protocol_specifies_update_dataset and $query->specifies_update_dataset) {
+				my $method	= $req->method;
+				$content	= "Update operations cannot specify a dataset in both the query and with protocol parameters";
+				$self->log_error( $req, $content );
+				$self->_set_response_error($req, $response, 400, {
+					title		=> "Multiple datasets specified for update",
+					describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/update_specifies_multiple_datasets',
+					detail		=> $content,
+				});
+				goto CLEANUP;
+			}
 			my ($plan, $ctx)	= $query->prepare( $model );
 # 			warn $plan->sse;
 			my $iter	= $query->execute_plan( $plan, $ctx );
@@ -268,11 +361,12 @@ END
 					$response->headers->header( ETag => $etag );
 				}
 				if ($iter->isa('RDF::Trine::Iterator::Graph')) {
-					my @variants	= (['text/html', 1.0, 'text/html']);
+					my @variants	= (['text/html', 0.99, 'text/html']);
 					my %media_types	= %RDF::Trine::Serializer::media_types;
 					while (my($type, $sclass) = each(%media_types)) {
 						next if ($type =~ /html/);
-						push(@variants, [$type, 0.99, $type]);
+						my $value	= ($type =~ m#application/rdf[+]xml#) ? 1.00 : 0.98;
+						push(@variants, [$type, $value, $type]);
 					}
 					my $stype	= choose( \@variants, $headers );
 					if ($stype !~ /html/ and my $sclass = $RDF::Trine::Serializer::media_types{ $stype }) {
@@ -316,20 +410,39 @@ END
 					}
 				}
 			} else {
-				$response->status(500);
-				$response->body($query->error);
+				my $error	= $query->error;
+				$self->_set_response_error($req, $response, 500, {
+					title		=> "SPARQL query/update execution error",
+					describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/execution_error',
+					detail		=> "$error; $sparql",
+				});
 				$content	= RDF::Query->error;
 			}
 		} else {
 			$content	= RDF::Query->error;
+			$self->log_error( $req, $content );
 			my $code	= ($content =~ /Syntax/) ? 400 : 500;
-			$response->status($code);
-			$response->body($content);
 			if ($req->method ne 'POST' and $content =~ /read-only queries/sm) {
 				$content	= 'Updates must use a HTTP POST request.';
+				$self->_set_response_error($req, $response, $code, {
+					title		=> $content,
+					describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/bad_http_method_update',
+				});
+			} else {
+				$self->_set_response_error($req, $response, $code, {
+					title		=> "SPARQL query/update parse error",
+					describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/parse_error',
+					detail		=> $content,
+				});
 			}
-			warn $content;
 		}
+	} elsif ($req->method eq 'POST') {
+		$content	= "POST without recognized query or update";
+		$self->log_error( $req, $content );
+		$self->_set_response_error($req, $response, 400, {
+			title		=> "Missing SPARQL Query/Update String",
+			describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/missing_sparql_string',
+		});
 	} else {
 		my @variants;
 		my %media_types	= %RDF::Trine::Serializer::media_types;
@@ -360,6 +473,10 @@ END
 		}
 	}
 	
+CLEANUP:
+# 	warn Dumper($model);
+# 	warn $model->as_string;
+	$content	= $response->body || $content;
 	my $length	= 0;
 	my %ae		= map { $_ => 1 } split(/\s*,\s*/, $ae);
 	if ($ae{'gzip'}) {
@@ -418,7 +535,10 @@ sub service_description {
 	}
 	
 	my $config		= $self->{conf};
+	my $doap		= RDF::Trine::Namespace->new('http://usefulinc.com/ns/doap#');
 	my $sd			= RDF::Trine::Namespace->new('http://www.w3.org/ns/sparql-service-description#');
+	my $sde			= RDF::Trine::Namespace->new('http://kasei.us/ns/service-description-extension#');
+	my $vann		= RDF::Trine::Namespace->new('http://purl.org/vocab/vann/');
 	my $void		= RDF::Trine::Namespace->new('http://rdfs.org/ns/void#');
 	my $scovo		= RDF::Trine::Namespace->new('http://purl.org/NET/scovo#');
 	my $count		= $model->count_statements( undef, undef, undef, RDF::Trine::Node::Nil->new );
@@ -445,13 +565,16 @@ sub service_description {
 	foreach my $func (@functions) {
 		$sdmodel->add_statement( statement( $s, $sd->extensionFunction, iri($func) ) );
 	}
+	
+	$sdmodel->add_statement( statement( $s, $sd->resultFormat, iri('http://www.w3.org/ns/formats/SPARQL_Results_XML') ) );
+	$sdmodel->add_statement( statement( $s, $sd->resultFormat, iri('http://www.w3.org/ns/formats/SPARQL_Results_JSON') ) );
 	foreach my $format (@formats) {
 		$sdmodel->add_statement( statement( $s, $sd->resultFormat, iri($format) ) );
 	}
 	
 	my $dataset		= blank('dataset');
 	$sdmodel->add_statement( statement( $s, $sd->endpoint, iri('') ) );
-	$sdmodel->add_statement( statement( $s, $sd->defaultDatasetDescription, $dataset ) );
+	$sdmodel->add_statement( statement( $s, $sd->defaultDataset, $dataset ) );
 	$sdmodel->add_statement( statement( $dataset, $rdf->type, $sd->Dataset ) );
 	if (my $d = $config->{endpoint}{service_description}{default}) {
 		my $def_graph	= ($d =~ /^\w+:/) ? iri($d) : blank('defaultGraph');
@@ -475,6 +598,25 @@ sub service_description {
 		}
 	}
 	
+	if (my $software = $config->{endpoint}{service_description}{software}) {
+		$sdmodel->add_statement( statement( $s, $sde->software, iri($software) ) );
+	}
+	
+	if (my $related = $config->{endpoint}{service_description}{related}) {
+		foreach my $r (@$related) {
+			$sdmodel->add_statement( statement( $s, $sde->relatedEndpoint, iri($r) ) );
+		}
+	}
+	
+	if (my $namespaces = $config->{endpoint}{service_description}{namespaces}) {
+		while (my($ns,$uri) = each(%$namespaces)) {
+			my $b	= RDF::Trine::Node::Blank->new();
+			$sdmodel->add_statement( statement( $s, $sde->namespace, $b ) );
+			$sdmodel->add_statement( statement( $b, $vann->preferredNamespacePrefix, literal($ns) ) );
+			$sdmodel->add_statement( statement( $b, $vann->preferredNamespaceUri, literal($uri) ) );
+		}
+	}
+	
 	$self->{ sd_cache }	= [ $etag, $sdmodel ];
 	return $sdmodel;
 }
@@ -496,7 +638,7 @@ sub iter_as_html {
 	my $html;
 
 	if (-r $file) {
-		open( my $fh, '<', $file ) or die $!;
+		open( my $fh, '<', $file ) or croak $!;
 		$html = do { local $/; <$fh>; };
 		close $fh;
 	} else {
@@ -639,6 +781,64 @@ sub node_as_html {
 	}
 }
 
+=item C<< log_query ( $message ) >>
+
+=cut
+
+sub log_query {
+	my $self	= shift;
+	my $req		= shift;
+	my $message	= shift;
+	$self->_log( $req, { level => 'info', message => $message } );
+}
+
+=item C<< log_error ( $message ) >>
+
+=cut
+
+sub log_error {
+	my $self	= shift;
+	my $req		= shift;
+	my $message	= shift;
+	$self->_log( $req, { level => 'error', message => $message } );
+}
+
+sub _log {
+	my $self	= shift;
+	my $req		= shift;
+	my $data	= shift;
+	my $logger	= $req->logger || sub {};
+	
+	$logger->($data);
+}
+
+sub _set_response_error {
+	my $self	= shift;
+	my $req		= shift;
+	my $resp	= shift;
+	my $code	= shift;
+	my $error	= shift;
+	my @variants	= (
+		['text/plain', 1.0, 'text/plain'],
+		['application/json-problem', 0.99, 'application/json-problem'],
+	);
+	my $headers	= $req->headers;
+	my $stype	= choose( \@variants, $headers ) || 'text/plain';
+	if ($stype eq 'application/json-problem') {
+		$resp->headers->content_type( 'application/json-problem' );
+		$resp->status($code);
+		my $content	= encode_json($error);
+		$resp->body($content);
+	} else {
+		$resp->headers->content_type( 'text/plain' );
+		$resp->status($code);
+		my @messages	= grep { defined($_) } @{ $error }{ qw(title detail) };
+		my $content		= join("\n\n", @messages);
+		$resp->body($content);
+	}
+	return;
+}
+
 =end private
 
 =cut
@@ -653,7 +853,7 @@ __END__
 
 =over 4
 
-=item * L<http://www.w3.org/TR/rdf-sparql-protocol/>
+=item * L<http://www.w3.org/TR/sparql11-protocol/>
 
 =item * L<http://www.perlrdf.org/>
 
@@ -669,7 +869,7 @@ __END__
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright (c) 2010 Gregory Todd Williams.
+Copyright (c) 2010-2014 Gregory Todd Williams.
 
 This software is provided 'as-is', without any express or implied
 warranty. In no event will the authors be held liable for any
